"""
This module contains the top-level routines for the quasisymmetric
stellarator construction.
"""

import logging
import numpy as np
from scipy.io import netcdf
import matplotlib.pyplot as plt

#logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

class Qic():
    """
    This is the main class for representing the quasisymmetric
    stellarator construction.
    """
    
    # Import methods that are defined in separate files:
    from .init_axis import init_axis, convert_to_spline
    from .calculate_r1 import _residual, _jacobian, solve_sigma_equation, \
        _determine_helicity, r1_diagnostics
    from .grad_B_tensor import calculate_grad_B_tensor, calculate_grad_grad_B_tensor, \
        Bfield_cylindrical, Bfield_cartesian, grad_B_tensor_cartesian, \
        grad_grad_B_tensor_cylindrical, grad_grad_B_tensor_cartesian
    from .calculate_r2 import calculate_r2, construct_qi_r2
    from .calculate_r3 import calculate_r3
    from .mercier import mercier
    from .plot import plot, get_boundary, B_fieldline, B_contour, plot_axis
    from .r_singularity import calculate_r_singularity
    from .plot import plot, plot_boundary, get_boundary, B_fieldline, B_contour, plot_axis, B_densityplot
    from .fourier_interpolation import fourier_interpolation
    from .Frenet_to_cylindrical import Frenet_to_cylindrical, to_RZ
    from .optimize_nae import optimise_params, min_geo_qi_consistency
    from .to_vmec import to_vmec
    from .util import B_mag
    
    def __init__(self, rc, zs, rs=[], zc=[], nfp=1, etabar=1., sigma0=0., B0=1.,
                 B0_vals=[], B0_svals=[], d_cvals=[], d_svals=[], alpha_cvals=[0.], alpha_svals=[0.], phi_shift=0,
                 omn = False, omn_method='buffer', k_buffer=1, p_buffer=2, delta  = np.pi/5, k_second_order_SS = 0, I2=0., sG=1, spsi=1, nphi=31,
                 B2s=0., B2c=0., B2s_cvals=[], B2s_svals=[], B2c_cvals=[], B2c_svals=[], p2=0., order="r1", d_over_curvature=0, d_over_curvature_cvals = []):
        """
        Create a near-axis stellarator.
        """
        # First, force {rc, zs, rs, zc} to have the same length, for
        # simplicity.
        nfourier = np.max([len(rc), len(zs), len(rs), len(zc)])
        self.nfourier = nfourier
        self.rc = np.zeros(nfourier)
        self.zs = np.zeros(nfourier)
        self.rs = np.zeros(nfourier)
        self.zc = np.zeros(nfourier)
        self.rc[:len(rc)] = rc
        self.zs[:len(zs)] = zs
        self.rs[:len(rs)] = rs
        self.zc[:len(zc)] = zc

        # Solve for omnigenity
        self.omn = omn
        self.delta  = min(abs(delta),0.95*np.pi/nfp)
        self.omn_method = omn_method
        self.k_buffer = k_buffer
        self.p_buffer = p_buffer
        self.k_second_order_SS = k_second_order_SS
        self.d_over_curvature = d_over_curvature
        self.d_over_curvature_cvals = d_over_curvature_cvals

        # Force nphi to be odd:
        if np.mod(nphi, 2) == 0:
            nphi += 1

        if sG != 1 and sG != -1:
            raise ValueError('sG must be +1 or -1')
        
        if spsi != 1 and spsi != -1:
            raise ValueError('spsi must be +1 or -1')

        self.nfp = nfp
        self.etabar = etabar
        self.sigma0 = sigma0
        phi = np.linspace(0, 2 * np.pi / nfp, nphi, endpoint=False)
        self.d_phi = phi[1] - phi[0]
        if omn==True and phi_shift==0:
            phi_shift=1/3
        self.phi_shift = phi_shift
        self.phi = phi + self.phi_shift*self.d_phi
        if B0_vals==[]:
            self.B0_vals = [B0]
            self.B0_well_depth = 0
        else:
            self.B0_vals = B0_vals
            self.B0_well_depth = B0_vals[1]
        self.B0 =  np.array(sum([self.B0_vals[i]*np.cos(nfp*i*self.phi) for i in range(len(self.B0_vals))]))
        self.B0_svals = B0_svals
        self.B0 += np.array(sum([self.B0_svals[i]*np.sin(nfp*i*self.phi) for i in range(len(self.B0_svals))]))
        if d_cvals==[] and d_svals == [] and d_over_curvature_cvals == [] and d_over_curvature == 0:
            self.d_cvals = [etabar]
            self.d_svals = d_svals
        else:
            self.d_cvals = d_cvals
            self.d_svals = d_svals
            if self.d_cvals == []:
                self.etabar = 0
                self.d_cvals = [0]
            else:
                self.etabar = d_cvals[0]
        if d_over_curvature_cvals == [] and not d_over_curvature == 0:
            self.d_over_curvature_cvals == [d_over_curvature]
        if not d_over_curvature_cvals == [] and not d_over_curvature == 0:
            raise ValueError('Only one of d_over_curvature_cvals or d_over_curvature should be specified.')
        self.d = np.array(sum([self.d_cvals[i]*np.cos(nfp*i*self.phi) for i in range(len(self.d_cvals))]))
        self.d = self.d + np.array(sum([self.d_svals[i]*np.sin(nfp*i*self.phi) for i in range(len(self.d_svals))]))
        self.alpha_cvals = alpha_cvals
        self.alpha_svals = alpha_svals
        self.alpha = np.array(sum([self.alpha_cvals[i]*np.cos(nfp*i*self.phi) for i in range(len(self.alpha_cvals))]))
        self.alpha = self.alpha + np.array(sum([self.alpha_svals[i]*np.sin(nfp*i*self.phi) for i in range(len(self.alpha_svals))]))
        self.B1s = self.B0 * self.d * np.sin(self.alpha)
        self.B1c = self.B0 * self.d * np.cos(self.alpha)
        self.I2 = I2
        self.sG = sG
        self.spsi = spsi
        self.nphi = nphi
        if B2c_cvals==[] and B2c_svals == []:
            self.B2c_cvals = [B2c]
            self.B2c_svals = B2c_svals
            self.B2c = B2c
        else:
            self.B2c_cvals = B2c_cvals
            self.B2c_svals = B2c_svals
            if self.B2c_cvals == []:
                self.B2c = 0
                self.B2c_cvals = [0]
            else:
                self.B2c = B2c_cvals[0]
        if B2s_cvals==[] and B2s_svals == []:
            self.B2s_cvals = [B2s]
            self.B2s_svals = B2s_svals
            self.B2s = B2s
        else:
            self.B2s_cvals = B2s_cvals
            self.B2s_svals = B2s_svals
            if self.B2s_cvals == []:
                self.B2s = 0
                self.B2s_cvals = [0]
            else:
                self.B2s = B2s_cvals[0]
        self.p2 = p2
        self.order = order
        self.min_R0_threshold = 0.3
        self.min_Z0_threshold = 0.3
        self._set_names()
        self.calculate()

    def change_nfourier(self, nfourier_new):
        """
        Resize the arrays of Fourier amplitudes. You can either increase
        or decrease nfourier.
        """
        rc_old = self.rc
        rs_old = self.rs
        zc_old = self.zc
        zs_old = self.zs
        index = np.min((self.nfourier, nfourier_new))
        self.rc = np.zeros(nfourier_new)
        self.rs = np.zeros(nfourier_new)
        self.zc = np.zeros(nfourier_new)
        self.zs = np.zeros(nfourier_new)
        self.rc[:index] = rc_old[:index]
        self.rs[:index] = rs_old[:index]
        self.zc[:index] = zc_old[:index]
        self.zs[:index] = zs_old[:index]
        nfourier_old = self.nfourier
        self.nfourier = nfourier_new
        self._set_names()
        # No need to recalculate if we increased the Fourier
        # resolution, only if we decreased it.
        self.calculate()

    def calculate(self):
        """
        Driver for the main calculations.
        """
        self.init_axis()
        self.solve_sigma_equation()
        self.r1_diagnostics()
        if self.order != 'r1':
            self.calculate_r2()
            if self.order == 'r3':
                self.calculate_r3()
    
    def get_dofs(self):
        """
        Return a 1D numpy vector of all possible optimizable
        degrees-of-freedom, for simsopt.
        """
        return np.concatenate((self.rc, self.zs, self.rs, self.zc,
                               np.array([self.etabar, self.sigma0, self.B2s, self.B2c, self.p2, self.I2, self.delta]),
                               self.B0_vals, self.B0_svals, self.d_cvals, self.d_svals, self.alpha_cvals, self.alpha_svals,
                               self.B2c_cvals, self.B2c_svals, self.B2s_cvals, self.B2s_svals, self.d_over_curvature_cvals,
                               np.array([self.k_second_order_SS,self.d_over_curvature])))

    def set_dofs(self, x):
        """
        For interaction with simsopt, set the optimizable degrees of
        freedom from a 1D numpy vector.
        """
        assert len(x) == self.nfourier * 4 + 7 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals)\
                       + len(self.alpha_cvals) + len(self.alpha_svals)\
                       + len(self.B2c_cvals) + len(self.B2c_svals) + len(self.B2s_cvals) + len(self.B2s_svals)\
                       + len(self.d_over_curvature_cvals) + 2
        self.rc = x[self.nfourier * 0 : self.nfourier * 1]
        self.zs = x[self.nfourier * 1 : self.nfourier * 2]
        self.rs = x[self.nfourier * 2 : self.nfourier * 3]
        self.zc = x[self.nfourier * 3 : self.nfourier * 4]
        self.etabar = x[self.nfourier * 4 + 0]
        self.sigma0 = x[self.nfourier * 4 + 1]
        self.B2s = x[self.nfourier * 4 + 2]
        self.B2c = x[self.nfourier * 4 + 3]
        self.p2 = x[self.nfourier * 4 + 4]
        self.I2 = x[self.nfourier * 4 + 5]
        self.delta = x[self.nfourier * 4 + 6]
        self.B0_vals  = x[self.nfourier * 4 + 6 + 1 : self.nfourier * 4 + 6 + len(self.B0_vals) + 1]
        self.B0_svals = x[self.nfourier * 4 + 6 + len(self.B0_vals) + 1 : self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + 1]
        self.d_cvals  = x[self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + 1 : self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + 1]
        self.d_svals  = x[self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + 1 : self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + 1]
        self.alpha_cvals = x[self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + 1 : self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + len(self.alpha_cvals) + 1]
        self.alpha_svals = x[self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + len(self.alpha_cvals) + 1 : self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + len(self.alpha_cvals) + len(self.alpha_svals) + 1]
        self.B2c_cvals   = x[self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + len(self.alpha_cvals) + len(self.alpha_svals) + 1 : self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + len(self.alpha_cvals) + len(self.alpha_svals) + len(self.B2c_cvals) + 1]
        self.B2c_svals   = x[self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + len(self.alpha_cvals) + len(self.alpha_svals) + len(self.B2c_cvals) + 1 : self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + len(self.alpha_cvals) + len(self.alpha_svals) + len(self.B2c_cvals) + len(self.B2c_svals) + 1]
        self.B2s_cvals   = x[self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + len(self.alpha_cvals) + len(self.alpha_svals) + len(self.B2c_cvals) + len(self.B2c_svals) + 1 : self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + len(self.alpha_cvals) + len(self.alpha_svals) + len(self.B2c_cvals) + len(self.B2c_svals) + len(self.B2s_cvals) + 1]
        self.B2s_svals               = x[self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + len(self.alpha_cvals) + len(self.alpha_svals) + len(self.B2c_cvals) + len(self.B2c_svals) + len(self.B2s_cvals) + 1 : self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + len(self.alpha_cvals) + len(self.alpha_svals) + len(self.B2c_cvals) + len(self.B2c_svals) + len(self.B2s_cvals) + len(self.B2s_svals) + 1]
        self.d_over_curvature_cvals  = x[self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + len(self.alpha_cvals) + len(self.alpha_svals) + len(self.B2c_cvals) + len(self.B2c_svals) + len(self.B2s_cvals) + len(self.B2s_svals) + 1 : self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + len(self.alpha_cvals) + len(self.alpha_svals) + len(self.B2c_cvals) + len(self.B2c_svals) + len(self.B2s_cvals) + len(self.B2s_svals) + len(self.d_over_curvature_cvals) + 1]
        self.k_second_order_SS       = x[self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + len(self.alpha_cvals) + len(self.alpha_svals) + len(self.B2c_cvals) + len(self.B2c_svals) + len(self.B2s_cvals) + len(self.B2s_svals) + len(self.d_over_curvature_cvals) + 1]
        self.d_over_curvature        = x[self.nfourier * 4 + 6 + len(self.B0_vals) + len(self.B0_svals) + len(self.d_cvals) + len(self.d_svals) + len(self.alpha_cvals) + len(self.alpha_svals) + len(self.B2c_cvals) + len(self.B2c_svals) + len(self.B2s_cvals) + len(self.B2s_svals) + len(self.d_over_curvature_cvals) + 2]
        # Set new B0, d, alpha, B1 and B2
        if len(self.B0_vals)>1:
            self.B0_well_depth = self.B0_vals[1]
        self.B0 = sum([self.B0_vals[i]*np.cos(self.nfp*i*self.phi) for i in range(len(self.B0_vals))])
        self.B0_svals = self.B0_svals
        self.B0 += np.array(sum([self.B0_svals[i]*np.sin(self.nfp*i*self.phi) for i in range(len(self.B0_svals))]))
        self.d  = np.array(sum([self.d_cvals[i]*np.cos(self.nfp*i*self.phi) for i in range(len(self.d_cvals))]))
        self.d += np.array(sum([self.d_svals[i]*np.sin(self.nfp*i*self.phi) for i in range(len(self.d_svals))]))
        self.alpha  = np.array(sum([self.alpha_cvals[i]*np.cos(self.nfp*i*self.phi) for i in range(len(self.alpha_cvals))]))
        self.alpha += np.array(sum([self.alpha_svals[i]*np.sin(self.nfp*i*self.phi) for i in range(len(self.alpha_svals))]))
        self.B1s = self.B0 * self.d * np.sin(self.alpha)
        self.B1c = self.B0 * self.d * np.cos(self.alpha)
        self.calculate()
        logger.info('set_dofs called with x={}. Now iota={}, elongation={}'.format(x, self.iota, self.max_elongation))
        
    def _set_names(self):
        """
        For simsopt, sets the list of names for each degree of freedom.
        """
        names = []
        names += ['rc({})'.format(j) for j in range(self.nfourier)]
        names += ['zs({})'.format(j) for j in range(self.nfourier)]
        names += ['rs({})'.format(j) for j in range(self.nfourier)]
        names += ['zc({})'.format(j) for j in range(self.nfourier)]
        names += ['etabar', 'sigma0', 'B2s', 'B2c', 'p2', 'I2', 'delta']
        names += ['B0({})'.format(j) for j in range(len(self.B0_vals))]
        names += ['B0s({})'.format(j) for j in range(len(self.B0_svals))]
        names += ['dc({})'.format(j) for j in range(len(self.d_cvals))]
        names += ['ds({})'.format(j) for j in range(len(self.d_svals))]
        names += ['alphac({})'.format(j) for j in range(len(self.alpha_cvals))]
        names += ['alphas({})'.format(j) for j in range(len(self.alpha_svals))]
        names += ['B2cc({})'.format(j) for j in range(len(self.B2c_cvals))]
        names += ['B2cs({})'.format(j) for j in range(len(self.B2c_svals))]
        names += ['B2sc({})'.format(j) for j in range(len(self.B2s_cvals))]
        names += ['B2ss({})'.format(j) for j in range(len(self.B2s_svals))]
        names += ['d_over_curvaturec({})'.format(j) for j in range(len(self.d_over_curvature_cvals))]
        names += ['k_second_order_SS','d_over_curvature']
        self.names = names

    @classmethod
    def from_paper(cls, name, **kwargs):
        """
        Get one of the configurations that has been used in our papers.
        Available values for ``name`` are
        ``"QI"``,
        ``"r1 section 5.1"``,
        ``"r1 section 5.2"``,
        ``"r1 section 5.3"``,
        ``"r2 section 5.1"``,
        ``"r2 section 5.2"``,
        ``"r2 section 5.3"``,
        ``"r2 section 5.4"``,
        ``"r2 section 5.5"``,
        These last 5 configurations can also be obtained by specifying an integer 1-5 for ``name``.
        The configurations that begin with ``"r1"`` refer to sections in 
        Landreman, Sengupta, and Plunk, Journal of Plasma Physics 85, 905850103 (2019).
        The configurations that begin with ``"r2"`` refer to sections in 
        Landreman and Sengupta, Journal of Plasma Physics 85, 815850601 (2019).
        The QI configuration refers to section 8.2 of
        Plunk, Landreman, and Helander, Journal of Plasma Physics 85, 905850602 (2019)

        You can specify any other arguments of the ``qic`` constructor
        in ``kwargs``. You can also use ``kwargs`` to override any of
        the properties of the configurations from the papers. For
        instance, you can modify the value of ``etabar`` in the first
        example using

        .. code-block::

          q = qic.Qic.from_paper('r1 section 5.1', etabar=1.1)
        """

        def add_default_args(kwargs_old, **kwargs_new):
            """
            Take any key-value arguments in ``kwargs_new`` and treat them as
            defaults, adding them to the dict ``kwargs_old`` only if
            they are not specified there.
            """
            for key in kwargs_new:
                if key not in kwargs_old:
                    kwargs_old[key] = kwargs_new[key]

                    
        if name == "r1 section 5.1":
            """ The configuration from Landreman, Sengupta, Plunk (2019), section 5.1 """
            add_default_args(kwargs, rc=[1, 0.045], zs=[0, -0.045], nfp=3, etabar=-0.9)
                
        elif name == "r1 section 5.2":
            """ The configuration from Landreman, Sengupta, Plunk (2019), section 5.2 """
            add_default_args(kwargs, rc=[1, 0.265], zs=[0, -0.21], nfp=4, etabar=-2.25)
                
        elif name == "r1 section 5.3":
            """ The configuration from Landreman, Sengupta, Plunk (2019), section 5.3 """
            add_default_args(kwargs, rc=[1, 0.042], zs=[0, -0.042], zc=[0, -0.025], nfp=3, etabar=-1.1, sigma0=-0.6)
                
        elif name == "r2 section 5.1" or name == '5.1' or name == 1:
            """ The configuration from Landreman & Sengupta (2019), section 5.1 """
            add_default_args(kwargs, rc=[1, 0.155, 0.0102], zs=[0, 0.154, 0.0111], nfp=2, etabar=0.64, order='r3', B2c=-0.00322)
            
        elif name == "r2 section 5.2" or name == '5.2' or name == 2:
            """ The configuration from Landreman & Sengupta (2019), section 5.2 """
            add_default_args(kwargs, rc=[1, 0.173, 0.0168, 0.00101], zs=[0, 0.159, 0.0165, 0.000985], nfp=2, etabar=0.632, order='r3', B2c=-0.158)
                             
        elif name == "r2 section 5.3" or name == '5.3' or name == 3:
            """ The configuration from Landreman & Sengupta (2019), section 5.3 """
            add_default_args(kwargs, rc=[1, 0.09], zs=[0, -0.09], nfp=2, etabar=0.95, I2=0.9, order='r3', B2c=-0.7, p2=-600000.)
                             
        elif name == "r2 section 5.4" or name == '5.4' or name == 4:
            """ The configuration from Landreman & Sengupta (2019), section 5.4 """
            add_default_args(kwargs, rc=[1, 0.17, 0.01804, 0.001409, 5.877e-05],
                       zs=[0, 0.1581, 0.01820, 0.001548, 7.772e-05], nfp=4, etabar=1.569, order='r3', B2c=0.1348)
                             
        elif name == "r2 section 5.5" or name == '5.5' or name == 5:
            """ The configuration from Landreman & Sengupta (2019), section 5.5 """
            add_default_args(kwargs, rc=[1, 0.3], zs=[0, 0.3], nfp=5, etabar=2.5, sigma0=0.3, I2=1.6, order='r3', B2c=1., B2s=3., p2=-0.5e7)
        elif name == "QI" or name == "QI r1 Plunk" or name == "QI Plunk":
            """ The configuration from Plunk, Landreman & Helander (2019), section 8.2 """
            rc      = [ 1.0, 0.0,-0.2 ]
            zs      = [ 0.0, 0.0, 0.35]
            B0_vals = [ 1.0, 0.1 ]
            d_svals = [ 0.0, 1.08, 0.26, 0.46]
            delta   = 0.1 * 2*np.pi
            nphi    = 151
            add_default_args(kwargs, rc=rc, zs=zs, nfp=1, B0_vals=B0_vals, d_svals=d_svals, nphi=nphi, omn=True, delta=delta)
        elif name == "QI r1 Jorge" or name == "QI NFP1 r1 Jorge" or name == "QI Jorge":
            """ The configuration from Jorge et al (2022) """
            rc      = [ 1.0,0.0,-0.4056622889934463,0.0,0.07747378220100756,0.0,-0.007803860877024245,0.0,0.0,0.0,0.0,0.0,0.0 ]
            zs      = [ 0.0,0.0,-0.24769666390049602,0.0,0.06767352436978152,0.0,-0.006980621303449165,0.0,-0.0006816270917189934,0.0,-1.4512784317099981e-05,0.0,-2.839050532138523e-06 ]
            B0_vals = [ 1.0,0.16915531046156507 ]
            omn_method ='non-zone'
            k_buffer = 3
            k_second_order_SS   = 0.0
            d_over_curvature   = 0.5183783762725197
            d_svals = [ 0.0,0.003563114185517955,0.0002015921485566435,-0.0012178616509882368,-0.00011629450296628697,-8.255825435616736e-07,3.2011540526397e-06 ]
            delta   = 0.1
            nfp     = 1
            nphi    = 201
            add_default_args(kwargs, omn_method = omn_method, k_buffer=k_buffer, rc=rc,zs=zs, nfp=nfp, B0_vals=B0_vals, d_svals=d_svals, nphi=nphi, omn=True, delta=delta, d_over_curvature=d_over_curvature, k_second_order_SS=k_second_order_SS)
        elif name == "QI NFP1 r2":
            rc      = [ 1.0,0.0,-0.41599809655680886,0.0,0.08291443961920232,0.0,-0.008906891641686355,0.0,0.0,0.0,0.0,0.0,0.0 ]
            zs      = [ 0.0,0.0,-0.28721210154364263,0.0,0.08425262593215394,0.0,-0.010427621520053335,0.0,-0.0008921610906627226,0.0,-6.357200965811029e-07,0.0,2.7316247301500753e-07 ]
            rs      = [ 0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0 ]
            zc      = [ 0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0 ]
            sigma0  =  0.0
            B0_vals = [ 1.0,0.15824229612567256 ]
            omn_method ='non-zone'
            k_buffer = 3
            p_buffer = 2
            k_second_order_SS   = 0.0
            d_over_curvature   = 0.48654821249917474
            d_svals = [ 0.0,-0.00023993050759319644,1.6644294162908823e-05,0.00012071143120099562,-1.1664837950174757e-05,-2.443821681789672e-05,2.0922298879435957e-06 ]
            delta   = 0.1
            nfp     = 1
            B2s_svals = [ 0.0,0.27368018673599265,-0.20986698715787325,0.048031641735420336,0.07269565329289157,1.3981498114634812e-07,-9.952017662433159e-10 ]
            B2c_cvals = [ -0.0007280714400220894,0.20739775852289746,0.05816363701644946,0.06465766308954603,0.006987357785313118,1.2229700694973357e-07,-3.057497440766065e-09,0.0 ]
            B2s_cvals = [ 0.0,0.0,0.0,0.0,0.0 ]
            B2c_svals = [ 0.0,0.0,0.0,0.0 ]
            p2      =  0.0
            nphi    = 201
            add_default_args(kwargs, sigma0 = sigma0, omn_method = omn_method, p_buffer = p_buffer, k_buffer=k_buffer, rs=rs,zc=zc, rc=rc,zs=zs, nfp=nfp, B0_vals=B0_vals, d_svals=d_svals, nphi=nphi, omn=True, delta=delta, B2c_cvals=B2c_cvals, B2s_svals=B2s_svals, p2=p2, order='r3', k_second_order_SS=k_second_order_SS, d_over_curvature=d_over_curvature, B2s_cvals=B2s_cvals, B2c_svals=B2c_svals)
        elif name == "QI NFP2 r2":
            rc      = [ 1.0,0.0,-0.07764451554933544,0.0,0.005284971468552636,0.0,-0.00016252676632564814,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0 ]
            zs      = [ 0.0,0.0,-0.06525233925323416,0.0,0.005858113288916291,0.0,-0.0001930489465183875,0.0,-1.21045713465733e-06,0.0,-6.6162738585035e-08,0.0,-1.8633251242689778e-07,0.0,1.4688345268925702e-07,0.0,-8.600467886165271e-08,0.0,4.172537468496238e-08,0.0,-1.099753830863863e-08 ]
            rs      = [ 0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0 ]
            zc      = [ 0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0 ]
            sigma0  =  0.0
            B0_vals = [ 1.0,0.12735237900304514 ]
            omn_method ='non-zone-fourier'
            k_buffer = 1
            p_buffer = 2
            k_second_order_SS   = -25.137439389881692
            d_over_curvature   = -0.14601620836497467
            d_svals = [ 0.0,-5.067489975338647,0.2759212337742016,-0.1407115065170644,0.00180521570352059,-0.03135134464554904,0.009582569807320895,-0.004780243312143034,0.002241790407060276,-0.0006738437017134619,0.00031559081192998053 ]
            delta   = 0.8
            nfp     = 2
            B2s_svals = [ 0.0,0.0012174780422017702,0.00026317725313621535,0.0002235661375254599,0.0006235230087895861,0.00021429298911807877,8.428032911991958e-05,-0.000142566391046771,-3.194627950185967e-05,-0.0001119389848119665,-6.226472957451552e-05 ]
            B2c_cvals = [ 0.0018400322140812674,-0.0013637739279265815,-0.0017961063281748597,-0.000855123667865997,-0.001412983361026517,-0.0010676686588779228,-0.0008117922713651492,-0.0002878689335032291,-0.0002515272886665927,-7.924709175875918e-05,-4.919421452969814e-05,0.0,0.0,0.0,0.0 ]
            B2s_cvals = [ 0.4445604502180231,0.13822067284200223,-0.561756934579829,0.2488873179399463,-0.14559282723014635,0.020548052084815048,-0.011070304464557718,0.004342889373034949,-0.0015730819049237866,0.0035406584522436986,0.002831887060104115,0.0,0.0,0.0,0.0 ]
            B2c_svals = [ 0.0,2.7062914673236698,-0.9151373916194634,0.021394010521077745,-0.017469913902854437,0.03186670312840335,0.021102584055813403,0.0024194864183551515,-0.0059152315287890125,0.003709416127750524,0.010027743000785166,0.0,0.0,0.0,0.0 ]
            p2      =  0.0
            nphi    = 201
            add_default_args(kwargs, sigma0 = sigma0, omn_method = omn_method, p_buffer = p_buffer, k_buffer=k_buffer, rs=rs,zc=zc, rc=rc,zs=zs, nfp=nfp, B0_vals=B0_vals, d_svals=d_svals, nphi=nphi, omn=True, delta=delta, B2c_cvals=B2c_cvals, B2s_svals=B2s_svals, p2=p2, order='r3', k_second_order_SS=k_second_order_SS, d_over_curvature=d_over_curvature, B2s_cvals=B2s_cvals, B2c_svals=B2c_svals)
        elif name == "LandremanPaul2021QA" or name == "precise QA":
            """
            A fit of the near-axis model to the quasi-axisymmetric
            configuration in Landreman & Paul, arXiv:2108.03711 (2021).

            The fit was performed to the boozmn data using the script
            20200621-01-Extract_B0_B1_B2_from_boozxform
            """
            add_default_args(kwargs,
                             nfp=2,
                             rc=[1.0038581971135636, 0.18400998741139907, 0.021723381370503204, 0.0025968236014410812, 0.00030601568477064874, 3.5540509760304384e-05, 4.102693907398271e-06, 5.154300428457222e-07, 4.8802742243232844e-08, 7.3011320375259876e-09],
                             zs=[0.0, -0.1581148860568176, -0.02060702320552523, -0.002558840496952667, -0.0003061368667524159, -3.600111450532304e-05, -4.174376962124085e-06, -4.557462755956434e-07, -8.173481495049928e-08, -3.732477282851326e-09],
                             B0=1.006541121335688,
                             etabar=-0.6783912804454629,
                             B2c=0.26859318908803137,
                             nphi=99,
                             order='r3')

        elif name == "precise QA+well":
            """
            A fit of the near-axis model to the precise quasi-axisymmetric
            configuration from SIMSOPT with magnetic well.

            The fit was performed to the boozmn data using the script
            20200621-01-Extract_B0_B1_B2_from_boozxform
            """
            add_default_args(kwargs,
                             nfp=2,
                             rc=[1.0145598919163676, 0.2106377247598754, 0.025469267136340394, 0.0026773601516136727, 0.00021104172568911153, 7.891887175655046e-06, -8.216044358250985e-07, -2.379942694112007e-07, -2.5495108673798585e-08, 1.1679227114962395e-08, 8.961288962248274e-09],
                             zs=[0.0, -0.14607192982551795, -0.021340448470388084, -0.002558983303282255, -0.0002355043952788449, -1.2752278964149462e-05, 3.673356209179739e-07, 9.261098628194352e-08, -7.976283362938471e-09, -4.4204430633540756e-08, -1.6019372369445714e-08],
                             B0=1.0117071561808106,
                             etabar=-0.5064143402495729,
                             B2c=-0.2749140163639202,
                             nphi=99,
                             order='r3')
            
        elif name == "LandremanPaul2021QH" or name == "precise QH":
            """
            A fit of the near-axis model to the quasi-helically symmetric
            configuration in Landreman & Paul, arXiv:2108.03711 (2021).

            The fit was performed to the boozmn data using the script
            20211001-02-Extract_B0_B1_B2_from_boozxform
            """
            add_default_args(kwargs,
                             nfp=4,
                             rc=[1.0033608429348413, 0.19993025252481125, 0.03142704185268144, 0.004672593645851904, 0.0005589954792333977, 3.298415996551805e-05, -7.337736061708705e-06, -2.8829857667619663e-06, -4.51059545517434e-07],
                             zs=[0.0, 0.1788824025525348, 0.028597666614604524, 0.004302393796260442, 0.0005283708386982674, 3.5146899855826326e-05, -5.907671188908183e-06, -2.3945326611145963e-06, -6.87509350019021e-07],
                             B0=1.003244143729638,
                             etabar=-1.5002839921360023,
                             B2c=0.37896407142157423,
                             nphi=99,
                             order='r3')

        elif name == "precise QH+well":
            """
            A fit of the near-axis model to the precise quasi-helically symmetric
            configuration from SIMSOPT with magnetic well.

            The fit was performed to the boozmn data using the script
            20211001-02-Extract_B0_B1_B2_from_boozxform
            """
            add_default_args(kwargs,
                             nfp=4,
                             rc=[1.000474932581454, 0.16345392520298313, 0.02176330066615466, 0.0023779201451133163, 0.00014141976024376502, -1.0595894482659743e-05, -2.9989267970578764e-06, 3.464574408947338e-08],
                             zs=[0.0, 0.12501739099323073, 0.019051257169780858, 0.0023674771227236587, 0.0001865909743321566, -2.2659053455802824e-06, -2.368335337174369e-06, -1.8521248561490157e-08],
                             B0=0.999440074325872,
                             etabar=-1.2115187546668142,
                             B2c=0.6916862277166693,
                             nphi=99,
                             order='r3')
        else:
            raise ValueError('Unrecognized configuration name')

        return cls(**kwargs)

    @classmethod
    def from_cxx(cls, filename):
        """
        Load a configuration from a ``qsc_out.<extension>.nc`` output file
        that was generated by the C++ version of QSC. Almost all the
        data will be taken from the output file, over-writing any
        calculations done in python when the new Qsc object is
        created.
        """
        def to_string(nc_str):
            """ Convert a string from the netcdf binary format to a python string. """
            temp = [c.decode('UTF-8') for c in nc_str]
            return (''.join(temp)).strip()
        
        f = netcdf.netcdf_file(filename, mmap=False)
        nfp = f.variables['nfp'][()]
        nphi = f.variables['nphi'][()]
        rc = f.variables['R0c'][()]
        rs = f.variables['R0s'][()]
        zc = f.variables['Z0c'][()]
        zs = f.variables['Z0s'][()]
        I2 = f.variables['I2'][()]
        B0 = f.variables['B0'][()]
        spsi = f.variables['spsi'][()]
        sG = f.variables['sG'][()]
        etabar = f.variables['eta_bar'][()]
        sigma0 = f.variables['sigma0'][()]
        order_r_option = to_string(f.variables['order_r_option'][()])
        if order_r_option == 'r2.1':
            order_r_option = 'r3'
        if order_r_option == 'r1':
            p2 = 0.0
            B2c = 0.0
            B2s = 0.0
        else:
            p2 = f.variables['p2'][()]
            B2c = f.variables['B2c'][()]
            B2s = f.variables['B2s'][()]

        q = cls(nfp=nfp, nphi=nphi, rc=rc, rs=rs, zc=zc, zs=zs,
                B0=B0, sG=sG, spsi=spsi,
                etabar=etabar, sigma0=sigma0, I2=I2, p2=p2, B2c=B2c, B2s=B2s, order=order_r_option)
        
        def read(name, cxx_name=None):
            if cxx_name is None: cxx_name = name
            setattr(q, name, f.variables[cxx_name][()])

        [read(v) for v in ['R0', 'Z0', 'R0p', 'Z0p', 'R0pp', 'Z0pp', 'R0ppp', 'Z0ppp',
                           'sigma', 'curvature', 'torsion', 'X1c', 'Y1c', 'Y1s', 'elongation']]
        if order_r_option != 'r1':
            [read(v) for v in ['X20', 'X2c', 'X2s', 'Y20', 'Y2c', 'Y2s', 'Z20', 'Z2c', 'Z2s', 'B20']]
            if order_r_option != 'r2':
                [read(v) for v in ['X3c1', 'Y3c1', 'Y3s1']]
                    
        f.close()
        return q
        
    def min_R0_penalty(self):
        """
        This function can be used in optimization to penalize situations
        in which min(R0) < min_R0_constraint.
        """
        return np.max((0, self.min_R0_threshold - self.min_R0)) ** 2

    def min_Z0_penalty(self):
        """
        This function can be used in optimization to penalize situations
        in which min(Z0) < min_Z0_constraint.
        """
        return np.max((0, self.min_Z0_threshold - self.min_Z0)) ** 2
        
    @classmethod
    def from_boozxform(cls, booz_xform_file, max_s_for_fit = 0.4, N_phi = 200, max_n_to_plot = 2, show=False,
                         vmec_file=None, rc=[], rs=[], zc=[], zs=[], nNormal=None, input_stel=None, savefig=False):#, order='r2', sigma0=0, I2=0, p2=0, omn=False):
        """
        Load a configuration from a VMEC and a BOOZ_XFORM output files
        """
        # Read properties of BOOZ_XFORM output file
        f = netcdf.netcdf_file(booz_xform_file,'r',mmap=False)
        bmnc = f.variables['bmnc_b'][()]
        ixm = f.variables['ixm_b'][()]
        ixn = f.variables['ixn_b'][()]
        jlist = f.variables['jlist'][()]
        ns = f.variables['ns_b'][()]
        nfp = f.variables['nfp_b'][()]
        Psi = f.variables['phi_b'][()]
        Psi_a = np.abs(Psi[-1])
        iotaVMECt=f.variables['iota_b'][()][1]
        f.close()

        if vmec_file!=None:
            # Read axis-shape from VMEC output file
            f = netcdf.netcdf_file(vmec_file,'r',mmap=False)
            am = f.variables['am'][()]
            rc = f.variables['raxis_cc'][()]
            zs = -f.variables['zaxis_cs'][()]
            try:
                rs = -f.variables['raxis_cs'][()]
                zc = f.variables['zaxis_cc'][()]
                logger.info('Non stellarator symmetric configuration')
            except:
                rs=[]
                zc=[]
                logger.info('Stellarator symmetric configuration')
            f.close()
        elif rc!=[]:
            # Read axis-shape from input parameters
            rc=rc
            rs=rs
            zc=zc
            zs=zs
        else:
            print("Axis shape not specified")
            # Calculate nNormal
        if nNormal==None:
            stel = Qic(rc=rc, rs=rs, zc=zc, zs=zs, nfp=nfp)
            nNormal = stel.iotaN - stel.iota
        else:
            nNormal = nNormal

        # Prepare coordinates for fit
        s_full = np.linspace(0,1,ns)
        ds = s_full[1] - s_full[0]
        #s_half = s_full[1:] - 0.5*ds
        s_half = s_full[jlist-1] - 0.5*ds
        mask = s_half < max_s_for_fit
        s_fine = np.linspace(0,1,400)
        sqrts_fine = s_fine
        phi = np.linspace(0,2*np.pi / nfp, N_phi)
        B0  = np.zeros(N_phi)
        B1s = np.zeros(N_phi)
        B1c = np.zeros(N_phi)
        B20 = np.zeros(N_phi)
        B2s = np.zeros(N_phi)
        B2c = np.zeros(N_phi)

        # Perform fit
        numRows=3
        numCols=max_n_to_plot*2+1
        if show: fig=plt.figure(num=None, figsize=(16, 9), dpi=80, facecolor='w', edgecolor='k')
        for jmn in range(len(ixm)):
            m = ixm[jmn]
            n = ixn[jmn] / nfp
            if m>2:
                continue
            doplot = (np.abs(n) <= max_n_to_plot) & show
            row = m
            col = n+max_n_to_plot
            if doplot:
                plt.subplot(int(numRows),int(numCols),int(row*numCols + col + 1))
                plt.plot(np.sqrt(s_half), bmnc[:,jmn],'.-')
                # plt.xlabel(r'$\sqrt{s}$')
                plt.title('bmnc(m='+str(m)+' n='+str(n)+')')
            if m==0:
                # For m=0, fit a polynomial in s (not sqrt(s)) that does not need to go through the origin.
                degree = 4
                p = np.polyfit(s_half[mask], bmnc[mask,jmn], degree)
                B0 += p[-1] * np.cos(n*nfp*phi)
                B20 += p[-2] * np.cos(n*nfp*phi)
                if doplot:
                    plt.plot(np.sqrt(s_fine), np.polyval(p, s_fine),'r')
            if m==1:
                # For m=1, fit a polynomial in sqrt(s) to an odd function
                x1 = np.sqrt(s_half[mask])
                y1 = bmnc[mask,jmn]
                x2 = np.concatenate((-x1,x1))
                y2 = np.concatenate((-y1,y1))
                degree = 5
                p = np.polyfit(x2,y2, degree)
                B1c += p[-2] * (np.sin(n*nfp*phi) * np.sin(nNormal*phi) + np.cos(n*nfp*phi) * np.cos(nNormal*phi))
                B1s += p[-2] * (np.sin(n*nfp*phi) * np.cos(nNormal*phi) - np.cos(n*nfp*phi) * np.sin(nNormal*phi))
                if doplot:
                    plt.plot(sqrts_fine, np.polyval(p, sqrts_fine),'r')
            if m==2:
                # For m=2, fit a polynomial in s (not sqrt(s)) that does need to go through the origin.
                x1 = s_half[mask]
                y1 = bmnc[mask,jmn]
                degree = 4
                p = np.polyfit(x1,y1, degree)
                B2c += p[-2] * (np.sin(n*nfp*phi) * np.sin(nNormal*phi) + np.cos(n*nfp*phi) * np.cos(nNormal*phi))
                B2s += p[-2] * (np.sin(n*nfp*phi) * np.cos(nNormal*phi) - np.cos(n*nfp*phi) * np.sin(nNormal*phi))
                if doplot:
                    plt.plot(np.sqrt(s_fine), np.polyval(p, s_fine),'r')
        if show:
            plt.show()
        # Convert expansion in sqrt(s) to an expansion in r
        BBar = np.mean(B0)
        sqrt_s_over_r = np.sqrt(np.pi * BBar / Psi_a)
        B1s *= sqrt_s_over_r
        B1c *= -sqrt_s_over_r
        B20 *= sqrt_s_over_r*sqrt_s_over_r
        B2c *= sqrt_s_over_r*sqrt_s_over_r
        B2s *= sqrt_s_over_r*sqrt_s_over_r
        eta_bar = np.mean(B1c) / BBar

        # NEEDS A WAY TO READ I2 FROM VMEC OR BOOZ_XFORM

        # if p2==0 and vmec_file!=None:
        #     r  = np.sqrt(Psi_a/(np.pi*BBar))
        #     p2 = - am[0]/r/r

        # if omn:
        #     if order=='r1':
        #         q = cls(rc=rc,rs=rs,zc=zc,zs=zs,etabar=eta_bar,nphi=N_phi,nfp=nfp,B0=BBar,sigma0=sigma0,I2=I2)
        #     else:
        #         q = cls(rc=rc,rs=rs,zc=zc,zs=zs,etabar=eta_bar,nphi=N_phi,nfp=nfp,B0=BBar,sigma0=sigma0,I2=I2, B2c=np.mean(B2c), B2s=np.mean(B2s), order=order, p2=p2)
        # else:
        #     if order=='r1':
        #         q = cls(rc=rc,rs=rs,zc=zc,zs=zs,etabar=eta_bar,nphi=N_phi,nfp=nfp,B0=BBar,sigma0=sigma0, I2=I2)
        #     else:
        #         q = cls(rc=rc,rs=rs,zc=zc,zs=zs,etabar=eta_bar,nphi=N_phi,nfp=nfp,B0=BBar,sigma0=sigma0, I2=I2, B2c=np.mean(B2c), B2s=np.mean(B2s), order=order, p2=p2)

        # q.B0_boozxform_array=B0
        # q.B1c_boozxform_array=B1c
        # q.B1s_boozxform_array=B1s
        # q.B20_boozxform_array=B20
        # q.B2c_boozxform_array=B2c
        # q.B2s_boozxform_array=B2s
        # q.iotaVMEC = iotaVMECt
        try:
            name = vmec_file[5:-3]

            figB0=plt.figure(figsize=(5, 5), dpi=80)
            plt.plot(input_stel.phi, input_stel.B0, 'r--', label=r'$B_0$ Near-axis')
            plt.plot(input_stel.phi, B0,            'b-' , label=r'$B_0$ VMEC')
            plt.xlabel(r'$\phi$', fontsize=18)
            plt.legend(fontsize=14)
            if savefig: figB0.savefig('B0_VMEC'+name+'.pdf')

            figB1=plt.figure(figsize=(5, 5), dpi=80)
            plt.plot(input_stel.phi, input_stel.B1c, 'r--', label=r'$B_{1c}$ Near-axis')
            plt.plot(input_stel.phi, B1c,            'r-' , label=r'$B_{1c}$ VMEC')
            plt.plot(input_stel.phi, input_stel.B1s, 'b--', label=r'$B_{1s}$ Near-axis')
            plt.plot(input_stel.phi, B1s,            'b-' , label=r'$B_{1s}$ VMEC')
            plt.xlabel(r'$\phi$', fontsize=18)
            plt.legend(fontsize=14)
            if savefig: figB1.savefig('B1_VMEC'+name+'.pdf')

            figB2=plt.figure(figsize=(5, 5), dpi=80)
            if input_stel.order != 'r1':
                plt.plot(input_stel.phi, input_stel.B20, 'r--', label=r'$B_{20}$ Near-axis')
                plt.plot(input_stel.phi, input_stel.B2c_array, 'b--', label=r'$B_{2c}$ Near-axis')
                plt.plot(input_stel.phi, input_stel.B2s_array, 'g--', label=r'$B_{2s}$ Near-axis')
            plt.plot(input_stel.phi, B20,            'r-' , label=r'$B_{20}$ VMEC')
            plt.plot(input_stel.phi, B2c,            'b-' , label=r'$B_{2c}$ VMEC')
            plt.plot(input_stel.phi, B2s,            'g-' , label=r'$B_{2s}$ VMEC')
            plt.xlabel(r'$\phi$', fontsize=18)
            plt.legend(fontsize=14)
            if savefig: figB2.savefig('B2_VMEC'+name+'.pdf')

            if show: plt.show()
        except Exception as e:
            print(e)

        plt.close(figB0)
        plt.close()

        return [B0,B1c,B1s,B20,B2c,B2s,iotaVMECt]


#!/usr/bin/env python3

import unittest
import numpy as np
from qsc.spectral_diff_matrix import spectral_diff_matrix

class SpectralDiffMatrixTests(unittest.TestCase):

    def test_Fourier(self):
        """
        Confirm that the derivative of a sin/cos is computed exactly.
        """
        xmin = -3.4
        xmax = -0.7
        L = xmax - xmin
        for nphi in range(11, 21, 2):
            D = spectral_diff_matrix(nphi, xmin=xmin, xmax=xmax)
            for n in range(0, int(np.floor(nphi/2)) - 1):
                for phase in [0, 0.3]:
                    phi = np.linspace(xmin, xmax, nphi, endpoint=False)
                    x = np.sin(n * phi * 2 * np.pi / L + phase)
                    dx = (n * 2 * np.pi / L) * np.cos(n * phi * 2 * np.pi / L + phase)
                    np.testing.assert_allclose(np.matmul(D, x), dx,
                                               rtol = 1e-12, atol = 1e-12)

    def test_random(self):
        """
        Do a bunch of tests for random xmin and xmax.
        """
        for n in range(1, 10):
            for whichrand in range(10):
                xmin = np.random.rand() * 20 - 10
                xmax = xmin + np.random.rand() * 20
                D = spectral_diff_matrix(n, xmin=xmin, xmax=xmax)
                # Check that the diagonal entries are all 0:
                for j in range(n):
                    self.assertLess(np.abs(D[j, j]), 1e-14)
                # Check that the row and column sums are all 0
                self.assertLess(np.max(np.abs(np.sum(D, axis=0))), 1e-11)
                self.assertLess(np.max(np.abs(np.sum(D, axis=1))), 1e-11)

    def test_n2(self):
        """
        Test matrix for n = 2
        """
        D1 = np.array([[0.0, 0.0], \
                       [0.0, 0.0]])
        D2 = spectral_diff_matrix(2)
        np.testing.assert_allclose(D1, D2, rtol=1e-13, atol=1e-13)
        
    def test_n3(self):
        """
        Test matrix for n = 3
        """
        x = 0.577350269189626
        D1 = np.array([[0, x, -x], \
                      [-x, 0, x], \
                      [x, -x, 0]])
        D2 = spectral_diff_matrix(3)
        np.testing.assert_allclose(D1, D2, rtol=1e-13, atol=1e-13)
        
    def test_n4(self):
        """
        Test matrix for n = 4
        """
        D1 = np.array([[0, 0.5, 0, -0.5], \
                       [-0.5, 0, 0.5, 0], \
                       [0, -0.5, 0, 0.5], \
                       [0.5, 0,-0.5, 0]])
        D2 = spectral_diff_matrix(4)
        np.testing.assert_allclose(D1, D2, rtol=1e-13, atol=1e-13)
        
    def test_n5(self):
        """
        Test matrix for n = 5
        """
        e = 0.85065080835204
        f = 0.525731112119134
        D1 = np.array([[0, e, -f, f, -e], \
                       [-e, 0, e, -f, f], \
                       [f, -e, 0, e, -f], \
                       [-f, f, -e, 0, e], \
                       [e, -f, f, -e, 0]])
        D2 = spectral_diff_matrix(5)
        np.testing.assert_allclose(D1, D2, rtol=1e-13, atol=1e-13)

    def test_n2_shifted(self):
        """
        Test matrix for n = 2, with xmin and xmax set.
        """
        D1 = np.array([[0.0, 0.0], \
                       [0.0, 0.0]])
        D2 = spectral_diff_matrix(2, xmin=-2.1, xmax=3.7)
        np.testing.assert_allclose(D1, D2, rtol=1e-13, atol=1e-13)
        
    def test_n3_shifted(self):
        """
        Test matrix for n = 3, with xmin and xmax set.
        """
        x = 0.625448056632489
        D1 = np.array([[0, x, -x], \
                      [-x, 0, x], \
                      [x, -x, 0]])
        D2 = spectral_diff_matrix(3, xmin=-2.1, xmax=3.7)
        np.testing.assert_allclose(D1, D2, rtol=1e-13, atol=1e-13)
        
    def test_n4_shifted(self):
        """
        Test matrix for n = 4, with xmin and xmax set.
        """
        x = 0.541653905791344
        D1 = np.array([[0, x, 0, -x], \
                       [-x, 0, x, 0], \
                       [0, -x, 0, x], \
                       [x, 0, -x, 0]])
        D2 = spectral_diff_matrix(4, xmin=-2.1, xmax=3.7)
        np.testing.assert_allclose(D1, D2, rtol=1e-13, atol=1e-13)
        
    def test_n5_shifted(self):
        """
        Test matrix for n = 5, with xmin and xmax set.
        """
        e = 0.921516665616892
        f = 0.569528620550711
        D1 = np.array([[0, e, -f, f, -e], \
                       [-e, 0, e, -f, f], \
                       [f, -e, 0, e, -f], \
                       [-f, f, -e, 0, e], \
                       [e, -f, f, -e, 0]])
        D2 = spectral_diff_matrix(5, xmin=-2.1, xmax=3.7)
        np.testing.assert_allclose(D1, D2, rtol=1e-13, atol=1e-13)
                
if __name__ == "__main__":
    unittest.main()



"""
This module contains the routine for computing the terms in
Mercier's criterion.
"""

import numpy as np
from .util import mu0

def mercier(self):
    """
    Compute the terms in Mercier's criterion.
    """

    # See Overleaf note "Mercier criterion near the magnetic axis- detailed notes".
    # See also "20200604-02 Checking sign in Mercier DGeod near axis.docx"

    # Shorthand:
    d_l_d_phi = self.d_l_d_phi
    B0 = self.B0
    Bbar = self.Bbar
    G0 = self.G0
    p2 = self.p2
    # etabar = self.etabar
    # curvature = self.curvature
    # sigma = self.sigma
    # iotaN = self.iotaN
    # iota = self.iota
    pi = np.pi
    d_phi = self.d_phi
    nfp = self.nfp
    B1s = self.B1s
    B1c = self.B1c
    B20 = self.B20
    X1s = self.X1s
    X1c = self.X1c
    Y1s = self.Y1s
    Y1c = self.Y1c
    beta_1s = self.beta_1s
    beta_1c = self.beta_1c

    #integrand = d_l_d_phi * (Y1c * Y1c + X1c * (X1c + Y1s)) / (Y1c * Y1c + (X1c + Y1s) * (X1c + Y1s))
    # integrand = d_l_d_phi * (etabar*etabar*etabar*etabar + curvature*curvature*curvature*curvature*sigma*sigma + etabar*etabar*curvature*curvature) \
    #     / (etabar*etabar*etabar*etabar + curvature*curvature*curvature*curvature*(1+sigma*sigma) + 2*etabar*etabar*curvature*curvature)

    # integral = np.sum(integrand) * self.d_phi * self.nfp * 2 * pi / self.axis_length

    #DGeod_times_r2 = -(2 * sG * spsi * mu0 * mu0 * p2 * p2 * G0 * G0 * G0 * G0 * etabar * etabar &
    # self.DGeod_times_r2 = -(2 * mu0 * mu0 * p2 * p2 * G0 * G0 * G0 * G0 * etabar * etabar \
    #                    / (pi * pi * pi * Bbar * Bbar * Bbar * Bbar * Bbar * Bbar * Bbar * Bbar * Bbar * Bbar * iotaN * iotaN)) \
    #                    * integral
    
    jacobian = self.sG * d_l_d_phi * B0 / G0

    V1 = X1s * X1s + X1c * X1c + Y1s * Y1s + Y1c * Y1c
    a = (X1s * X1s - X1c * X1c + Y1s * Y1s - Y1c * Y1c) / V1
    b = -2 * (X1s * X1c + Y1s * Y1c) / V1
    integrad_I_beta = (1/(B0 * B0 * V1)) * ((a * a + b * b)*(beta_1s * beta_1s + beta_1c * beta_1c) + \
                                             (np.sqrt(1 - a * a - b * b) - 1) * (a * (beta_1s * beta_1s - beta_1c * beta_1c) - 2 * b * beta_1s * beta_1c)) \
                                            / (np.sqrt(1 - a * a - b * b) * (a * a + b * b))
    I_beta  = np.sum(integrad_I_beta * jacobian) * d_phi * nfp
    self.DGeod_times_r2 = - abs(G0) * self.axis_length * I_beta / (16 * pi * pi * pi * pi * abs(Bbar))

    # self.d2_volume_d_psi2 = 4*pi*pi*abs(G0)/(Bbar*Bbar*Bbar)*(3*etabar*etabar - 4*self.B20_mean/Bbar + 2 * (self.G2 + iota * self.I2)/G0)
    integrand1 = 1 / (B0 * B0)
    integral1  = np.sum(integrand1 * jacobian) * d_phi * nfp
    integrand2 = (1 / (B0 * B0 * B0 * B0)) * (3 * (B1s * B1s + B1c * B1c) - 4 * B0 * B20 - mu0 * p2 * B0 * B0 / np.pi * integral1)
    integral2  = np.sum(integrand2 * jacobian) * d_phi * nfp
    self.d2_volume_d_psi2 = 2 * pi * abs(G0 / Bbar) * integral2

    # self.DWell_times_r2   = (mu0 * p2 * abs(G0) / (8 * pi * pi * pi * pi * Bbar * Bbar * Bbar)) * \
    #     (self.d2_volume_d_psi2 - 8 * pi * pi * mu0 * p2 * abs(G0) / (Bbar * Bbar * Bbar * Bbar * Bbar))
    self.DWell_times_r2 = mu0 * p2 * self.axis_length / (16 * pi * pi * pi * pi * pi * Bbar * Bbar) * \
        (self.d2_volume_d_psi2 - 4 * pi * mu0 * p2 * abs(G0 / Bbar) * np.sum((1/(B0 * B0 * B0 * B0)) * jacobian) * d_phi * nfp)

    self.DMerc_times_r2 = self.DWell_times_r2 + self.DGeod_times_r2



#!/usr/bin/env python3

"""
Various utility functions
"""

import logging
import numpy as np
import scipy.optimize
from qic.fourier_interpolation import fourier_interpolation
from scipy.interpolate import CubicSpline as spline

#logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

mu0 = 4 * np.pi * 1e-7

class Struct():
    """
    This class is just an empty mutable object to which we can attach
    attributes.
    """
    pass

def fourier_minimum(y):
    """
    Given uniformly spaced data y on a periodic domain, find the
    minimum of the spectral interpolant.
    """
    # Handle the case of a constant:
    if (np.max(y) - np.min(y)) / np.max([1e-14, np.abs(np.mean(y))]) < 1e-14:
        return y[0]
    
    n = len(y)
    dx = 2 * np.pi / n
    # Compute a rough guess for the minimum, given by the minimum of
    # the discrete data:
    index = np.argmin(y)

    def func(x):
        interp = fourier_interpolation(y, np.array([x]))
        logger.debug('fourier_minimum.func called at x={}, y={}'.format(x, interp[0]))
        return interp[0]

    # Try to find a bracketing interval, using successively wider
    # intervals.
    f0 = func(index * dx)
    found_bracket = False
    for j in range(1, 4):
        bracket = np.array([index - j, index, index + j]) * dx
        fm = func(bracket[0])
        fp = func(bracket[2])
        if f0 < fm and f0 < fp:
            found_bracket = True
            break
    if not found_bracket:
        # We could throw an exception, though scipy will do that anyway
        pass

    logger.info('bracket={}, f(bracket)={}'.format(bracket, [func(bracket[0]), func(bracket[1]), func(bracket[2])]))
    #solution = scipy.optimize.minimize_scalar(func, bracket=bracket, options={"disp": True})
    solution = scipy.optimize.minimize_scalar(func, bracket=bracket)
    return solution.fun

def to_Fourier(R_2D, Z_2D, nfp, mpol, ntor, lasym):
    """
    This function takes two 2D arrays (R_2D and Z_2D), which contain
    the values of the radius R and vertical coordinate Z in cylindrical
    coordinates of a given surface and Fourier transform it, outputing
    the resulting cos(theta) and sin(theta) Fourier coefficients

    The first dimension of R_2D and Z_2D should correspond to the
    theta grid, while the second dimension should correspond to the
    phi grid.

    Args:
        R_2D: 2D array of the radial coordinate R(theta, phi) of a given surface
        Z_2D: 2D array of the vertical coordinate Z(theta, phi) of a given surface
        nfp: number of field periods of the surface
        mpol: resolution in poloidal Fourier space
        ntor: resolution in toroidal Fourier space
        lasym: False if stellarator-symmetric, True if not
    """
    shape = np.array(R_2D).shape
    ntheta = shape[0]
    nphi_conversion = shape[1]
    theta = np.linspace(0, 2 * np.pi, ntheta, endpoint=False)
    phi_conversion = np.linspace(0, 2 * np.pi / nfp, nphi_conversion, endpoint=False)
    RBC = np.zeros((int(2 * ntor + 1), int(mpol + 1)))
    RBS = np.zeros((int(2 * ntor + 1), int(mpol + 1)))
    ZBC = np.zeros((int(2 * ntor + 1), int(mpol + 1)))
    ZBS = np.zeros((int(2 * ntor + 1), int(mpol + 1)))
    factor = 2 / (ntheta * nphi_conversion)
    phi2d, theta2d = np.meshgrid(phi_conversion, theta)
    for m in range(mpol+1):
        nmin = -ntor
        if m==0: nmin = 1
        for n in range(nmin, ntor+1):
            angle = m * theta2d - n * nfp * phi2d
            sinangle = np.sin(angle)
            cosangle = np.cos(angle)
            factor2 = factor
            # The next 2 lines ensure inverse Fourier transform(Fourier transform) = identity
            if np.mod(ntheta,2) == 0 and m  == (ntheta/2): factor2 = factor2 / 2
            if np.mod(nphi_conversion,2) == 0 and abs(n) == (nphi_conversion/2): factor2 = factor2 / 2
            RBC[n + ntor, m] = np.sum(R_2D * cosangle * factor2)
            RBS[n + ntor, m] = np.sum(R_2D * sinangle * factor2)
            ZBC[n + ntor, m] = np.sum(Z_2D * cosangle * factor2)
            ZBS[n + ntor, m] = np.sum(Z_2D * sinangle * factor2)
    RBC[ntor,0] = np.sum(R_2D) / (ntheta * nphi_conversion)
    ZBC[ntor,0] = np.sum(Z_2D) / (ntheta * nphi_conversion)

    if not lasym:
        RBS = 0
        ZBC = 0

    return RBC, RBS, ZBC, ZBS

def B_mag(self, r, theta, phi, Boozer_toroidal = False, B0=1):
    '''
    Function to calculate the modulus of the magnetic field B for a given
    near-axis radius r, a Boozer poloidal angle theta (not vartheta) and
    a cylindrical toroidal angle phi if Boozer_toroidal = True or the
    Boozer angle varphi if Boozer_toroidal = True

    Args:
      r: the near-axis radius
      theta: the Boozer poloidal angle
      phi: the cylindrical or Boozer toroidal angle
      Boozer_toroidal: False if phi is the cylindrical toroidal angle, True for the Boozer one
    '''
    if Boozer_toroidal == False:
        varphi = phi + self.nu_spline(phi)
    else:
        varphi = phi
        phi    = varphi - self.nu_spline_of_varphi(varphi)

    thetaN = theta-self.N_helicity*varphi

    if self.order == 'r1':
        return B0*self.B0_spline(phi) + r * (self.B1c_spline(phi) * np.cos(thetaN) + self.B1s_spline(phi) * np.sin(thetaN))
    else:
        return B0*self.B0_spline(phi) + r * (self.B1c_spline(phi) * np.cos(thetaN) + self.B1s_spline(phi) * np.sin(thetaN)) \
                                   + r * r * (self.B20_spline(phi) + self.B2c_spline(phi) * np.cos(2*thetaN) + self.B2s_spline(phi) * np.sin(2*thetaN))


#!/usr/bin/env python3

import unittest
import numpy as np
from qsc.util import *

class FourierMinimizeTests(unittest.TestCase):

    def test_sin(self):
        """
        If data represents a single Fourier mode, we know the minimum analytically
        """
        for ndata in [4, 11]:
            for const in [-0.4, 1.3]:
                for amplitude in [2.2, 3.7]:
                    x = np.linspace(0, 2 * np.pi, ndata, endpoint=False)
                    y = const + amplitude * np.sin(x - 1.7)
                    min = fourier_minimum(y)
                    self.assertAlmostEqual(min, const - amplitude, places=14)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

"""
This module contains a subroutine for spectrally accurate interpolation of
data that is known on a uniform grid in a periodic domain.
"""

import numpy as np
#from numba import njit

# Get machine precision
eps = np.finfo(float).eps
    
#@njit
def fourier_interpolation(fk, x):
    """
    Interpolate data that is known on a uniform grid in [0, 2pi).

    This routine is based on the
    matlab routine fourint.m in the DMSuite package by S.C. Reddy and J.A.C. Weideman, available at
    http://www.mathworks.com/matlabcentral/fileexchange/29
    or here:
    http://dip.sun.ac.za/~weideman/research/differ.html  

    fk:  Vector of y-coordinates of data, at equidistant points
         x(k) = (k-1)*2*pi/N,  k = 1...N
    x:   Vector of x-values where interpolant is to be evaluated.
    output: Vector of interpolated values.
    """

    N = len(fk)
    M = len(x)

    # Compute equidistant points
    #xk = np.linspace(0.0, 2 * np.pi, N, endpoint=False)
    xk = (np.arange(N) * 2 * np.pi) / N

    # Weights for trig interpolation
    w = (-1.0) ** np.arange(0, N)
    #w = np.array((-1) ** np.arange(0, N), dtype='f')

    """
    x2 = x / 2
    xk2 = xk / 2

    # Compute quantities x - x(k)
    xk2_2D, x2_2D = np.meshgrid(xk2, x2)
    Dold = x2_2D - xk2_2D
    D = 0.5 * (np.outer(x, np.ones(N)) - np.outer(np.ones(M), xk))
    print(Dold - D)
    """
    D = 0.5 * (np.outer(x, np.ones(N)) - np.outer(np.ones(M), xk))
    
    if np.mod(N, 2) == 0:
        # Formula for N even
        D = 1 / np.tan(D + eps * (D==0))
    else:
        # Formula for N odd
        D = 1 / np.sin(D + eps * (D==0))

    # Evaluate interpolant as matrix-vector products
    #return np.matmul(D, w * fk) / np.matmul(D, w)
    return np.dot(D, w * fk) / np.dot(D, w)
    #return (D @ w * fk) / (D @ w)
    #return D.dot(w * fk) / D.dot(w)

def fourier_interpolation_matrix(N, x):
    x = np.array([x])
    M = len(x)
    xk = (np.arange(N) * 2 * np.pi) / N
    w = (-1.0) ** np.arange(0, N)
    x2 = x / 2
    xk2 = xk / 2
    xk2_2D, x2_2D = np.meshgrid(xk2, x2)
    D = x2_2D - xk2_2D
    if np.mod(N, 2) == 0:
        D = 1 / np.tan(D + eps * (D==0))
    else:
        D = 1 / np.sin(D + eps * (D==0))
    return np.diag(1/(D @ w)) @ D @ np.diag(w)



#!/usr/bin/env python3

import unittest
import numpy as np
import logging
from qsc.qsc import Qsc

logger = logging.getLogger(__name__)

class GradBTensorTests(unittest.TestCase):

    def a_test_axisymmetry(self):
        """
        Test that the grad B tensor is correct for an axisymmetric vacuum field
        """
        for sG in [-1, 1]:
            for spsi in [-1, 1]:
                for nfp in range(1, 4):
                    for j in range(3):
                        R0 = np.random.rand() * 2 + 0.3
                        B0 = np.random.rand() * 2 + 0.3
                        nphi = int(np.random.rand() * 10 + 3)
                        etabar = (np.random.rand() - 0.5) * 4
                        stel = Qsc(rc=[R0], zs=[0], nfp=nfp, nphi=nphi, sG=sG, \
                                   spsi=spsi, B0=B0, etabar=etabar)
                        val = sG * B0 / R0
                        if np.mod(nphi, 2) == 0:
                            nphi += 1

                        rtol = 1e-13
                        atol = 1e-13
                        np.testing.assert_allclose(stel.grad_B_tensor.nt, np.full(nphi, val), rtol=rtol, atol=atol)
                        np.testing.assert_allclose(stel.grad_B_tensor.tn, np.full(nphi, val), rtol=rtol, atol=atol)
                        np.testing.assert_allclose(stel.grad_B_tensor.nn, np.zeros(nphi), rtol=rtol, atol=atol)
                        np.testing.assert_allclose(stel.grad_B_tensor.bb, np.zeros(nphi), rtol=rtol, atol=atol)
                        np.testing.assert_allclose(stel.grad_B_tensor.nb, np.zeros(nphi), rtol=rtol, atol=atol)
                        np.testing.assert_allclose(stel.grad_B_tensor.bn, np.zeros(nphi), rtol=rtol, atol=atol)
                        np.testing.assert_allclose(stel.L_grad_B, np.full(nphi, R0), rtol=rtol, atol=atol)
                        
class GradGradBTensorTests(unittest.TestCase):

    def a_test_alternate_derivation_and_symmetry(self):
        """
        Compare Rogerio's derivation to mine to ensure the results
        coincide.  Also verify symmetry in the first two components.
        """
        for sG in [-1, 1]:
            for spsi in [-1, 1]:
                for config in [1, 2, 3, 4, 5]:
                    B0 = np.random.rand() * 0.4 + 0.8
                    nphi = int(np.random.rand() * 50 + 61)
                    s = Qsc.from_paper(config, sG=sG, spsi=spsi, B0=B0, nphi=65)
                    s.calculate_grad_grad_B_tensor(two_ways=True)
                    logger.info("Max difference between Matt and Rogerio's derivation for config {} is {}".format(config, np.max(np.abs(s.grad_grad_B - s.grad_grad_B_alt))))
                    np.testing.assert_allclose(s.grad_grad_B, s.grad_grad_B_alt, rtol=1e-7, atol=1e-7)
                    for i in range(3):
                        for j in range(3):
                            for k in range(3):
                                # For all configs, the tensor should be symmetric in the 1st 2 indices:
                                np.testing.assert_allclose(s.grad_grad_B[:, i, j, k], s.grad_grad_B[:, j, i, k], rtol=1e-8, atol=1e-8)
                                # For curl-free fields, the tensor should also be symmetric in the last 2 indices:
                                if config in {1, 2, 4}:
                                    np.testing.assert_allclose(s.grad_grad_B[:, i, j, k], s.grad_grad_B[:, i, k, j], rtol=1e-8, atol=1e-8)
            
    def a_test_axisymmetry(self):
        """
        Test that the grad grad B tensor is correct for an axisymmetric vacuum field
        """
        B0_vals = [0.9, 1.1]
        R0_vals = [0.85, 1.0, 1.2]
        nphi_vals = [3, 5, 7, 9, 11]
        etabar_vals = [0.55, 0.65, 0.95, 1.0, 1.09, 1.3, 1.4]
        index = 0
        for sG in [-1, 1]:
            for spsi in [-1, 1]:
                for nfp in range(1, 4):
                    for j in range(3):
                        """
                        R0 = np.random.rand() * 2 + 0.3
                        B0 = np.random.rand() * 2 + 0.3
                        nphi = int(np.random.rand() * 10 + 3)
                        etabar = (np.random.rand() - 0.5) * 4
                        """
                        # Try many combinations of values, in a deterministic way:
                        R0 = R0_vals[np.mod(index, len(R0_vals))]
                        B0 = B0_vals[np.mod(index, len(B0_vals))]
                        nphi = nphi_vals[np.mod(index, len(nphi_vals))]
                        etabar = etabar_vals[np.mod(index, len(etabar_vals))]
                        index += 1
                        
                        s = Qsc(rc=[R0], zs=[0], nfp=nfp, nphi=nphi, sG=sG, \
                                   spsi=spsi, B0=B0, etabar=etabar, I2=1.e-8, order='r2')
                        # The O(r^2) solve fails if iota is exactly 0,
                        # so above we set a tiny nonzero current to
                        # create a tiny nonzero iota.
                        if np.mod(nphi, 2) == 0:
                            nphi += 1

                        s.calculate_grad_grad_B_tensor(two_ways=True)
                        # Check all components other than {tnn, ntn, nnt, ttt}. These should be 0.
                        rtol = 1e-6
                        atol = 1e-6
                        for i in range(3):
                            for j in range(3):
                                for k in range(3):
                                    if not ((i==2 and j==2 and k==2) or (i==2 and j==0 and k==0) or (i==0 and j==2 and k==0) or (i==i and j==0 and k==2)):
                                        np.testing.assert_allclose(s.grad_grad_B[:,i,j,k], np.zeros(nphi), rtol=rtol, atol=atol)
                                        np.testing.assert_allclose(s.grad_grad_B_alt[:,i,j,k], np.zeros(nphi), rtol=rtol, atol=atol)
                                        
                        # Check ttt component
                        val = -2 * sG * B0 / (R0 * R0)
                        np.testing.assert_allclose(s.grad_grad_B[:,2,2,2], np.full(nphi, val), rtol=rtol, atol=atol)
                        np.testing.assert_allclose(s.grad_grad_B_alt[:,2,2,2], np.full(nphi, val), rtol=rtol, atol=atol)

                        # Check tnn, ntn, and nnt components:
                        val = 2 * sG * B0 / (R0 * R0)
                        np.testing.assert_allclose(s.grad_grad_B[:,2,0,0], np.full(nphi, val), rtol=rtol, atol=atol)
                        np.testing.assert_allclose(s.grad_grad_B[:,0,2,0], np.full(nphi, val), rtol=rtol, atol=atol)
                        np.testing.assert_allclose(s.grad_grad_B[:,0,0,2], np.full(nphi, val), rtol=rtol, atol=atol)
                        np.testing.assert_allclose(s.grad_grad_B_alt[:,2,0,0], np.full(nphi, val), rtol=rtol, atol=atol)
                        np.testing.assert_allclose(s.grad_grad_B_alt[:,0,2,0], np.full(nphi, val), rtol=rtol, atol=atol)
                        np.testing.assert_allclose(s.grad_grad_B_alt[:,0,0,2], np.full(nphi, val), rtol=rtol, atol=atol)
                        

class CylindricalCartesianTensorsTests(unittest.TestCase):

    def test_cylindrical_cartesian_tensors(self):
        """
        Test the accuracy and the symmetry in the the cylindrical and Cartesian
        versions of the grad B and grad grad B tensors.
        """
        # Test the grad B tensor in axisymmetry, including symmetry in both indices
        Rmajor = np.random.rand() * 2 + 0.3
        B0 = np.random.rand() * 2 + 1.0
        stel = Qsc(rc=[Rmajor, 0.], zs=[0, 0.], etabar=1.0, B0=B0)
        factor = stel.B0 / Rmajor
        dBdx_cylindrical = stel.grad_B_tensor_cylindrical
        np.testing.assert_almost_equal(dBdx_cylindrical[0, 0], np.zeros(stel.nphi))
        np.testing.assert_almost_equal(dBdx_cylindrical[0, 1], np.full(stel.nphi, -factor))
        np.testing.assert_almost_equal(dBdx_cylindrical[1, 1], np.zeros(stel.nphi))
        np.testing.assert_almost_equal(dBdx_cylindrical, dBdx_cylindrical.transpose(1, 0, 2))
        dBdx_cartesian = stel.grad_B_tensor_cartesian()
        np.testing.assert_almost_equal(dBdx_cartesian[0, 0], np.full(stel.nphi, factor * np.sin(2 * stel.phi)))
        np.testing.assert_almost_equal(dBdx_cartesian[0, 1], np.full(stel.nphi, -factor * np.cos(2 * stel.phi)))
        np.testing.assert_almost_equal(dBdx_cartesian[1, 1], np.full(stel.nphi, -factor * np.sin(2 * stel.phi)))
        np.testing.assert_almost_equal(dBdx_cartesian, dBdx_cartesian.transpose(1, 0, 2))

        # Test the grad B tensor for another configuration, including symmetry in both indices
        stel = Qsc.from_paper(1)
        dBdx_cylindrical = stel.grad_B_tensor_cylindrical
        np.testing.assert_almost_equal(dBdx_cylindrical, dBdx_cylindrical.transpose(1, 0, 2))
        dBdx_cartesian = stel.grad_B_tensor_cartesian()
        np.testing.assert_almost_equal(dBdx_cartesian, dBdx_cartesian.transpose(1, 0, 2))

        # Test the grad grad B tensor, including symmetry in three indices
        dBdxdx_cylindrical = stel.grad_grad_B_tensor_cylindrical()
        np.testing.assert_almost_equal(np.transpose(stel.grad_grad_B, (1, 2, 3, 0)), dBdxdx_cylindrical)
        dBdxdx_cylindrical_transpose_1 = dBdxdx_cylindrical.transpose(0, 2, 1, 3)
        dBdxdx_cylindrical_transpose_2 = dBdxdx_cylindrical.transpose(1, 0, 2, 3)
        dBdxdx_cylindrical_transpose_3 = dBdxdx_cylindrical.transpose(1, 2, 0, 3)
        dBdxdx_cylindrical_transpose_4 = dBdxdx_cylindrical.transpose(2, 0, 1, 3)
        dBdxdx_cylindrical_transpose_5 = dBdxdx_cylindrical.transpose(2, 1, 0, 3)
        np.testing.assert_almost_equal(dBdxdx_cylindrical_transpose_1, dBdxdx_cylindrical)
        np.testing.assert_almost_equal(dBdxdx_cylindrical_transpose_2, dBdxdx_cylindrical)
        np.testing.assert_almost_equal(dBdxdx_cylindrical_transpose_3, dBdxdx_cylindrical)
        np.testing.assert_almost_equal(dBdxdx_cylindrical_transpose_4, dBdxdx_cylindrical)
        np.testing.assert_almost_equal(dBdxdx_cylindrical_transpose_5, dBdxdx_cylindrical)

        dBdxdx_cartesian = stel.grad_grad_B_tensor_cartesian()
        dBdxdx_cartesian_transpose_1 = dBdxdx_cartesian.transpose(0, 2, 1, 3)
        dBdxdx_cartesian_transpose_2 = dBdxdx_cartesian.transpose(1, 0, 2, 3)
        dBdxdx_cartesian_transpose_3 = dBdxdx_cartesian.transpose(1, 2, 0, 3)
        dBdxdx_cartesian_transpose_4 = dBdxdx_cartesian.transpose(2, 0, 1, 3)
        dBdxdx_cartesian_transpose_5 = dBdxdx_cartesian.transpose(2, 1, 0, 3)
        np.testing.assert_almost_equal(dBdxdx_cartesian_transpose_1, dBdxdx_cartesian)
        np.testing.assert_almost_equal(dBdxdx_cartesian_transpose_2, dBdxdx_cartesian)
        np.testing.assert_almost_equal(dBdxdx_cartesian_transpose_3, dBdxdx_cartesian)
        np.testing.assert_almost_equal(dBdxdx_cartesian_transpose_4, dBdxdx_cartesian)
        np.testing.assert_almost_equal(dBdxdx_cartesian_transpose_5, dBdxdx_cartesian)
                
if __name__ == "__main__":
    unittest.main()



"""
This module contains calculations for O(r^3) terms.
"""

import logging
import numpy as np
from .util import mu0

#logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

def calculate_r3(self):
    """
    Compute the O(r^3) contributions to X and Y needed for full
    consistency of B through O(r^2), as detailed in section 3 of
    Landreman & Sengupta, J Plasma Physics (2019).  This subroutine is
    based on the order_r_option = "r3_flux_constraint" equations from
    the fortran quasisymmetry code.
    """
    logger.debug('Calculating O(r^3) terms')
 
    # Shorthand
    sign_psi = self.spsi
    sign_G   = self.sG
    G2       = self.G2
    N_helicity = self.iota - self.iotaN
    B0 = self.B0
    Bbar = self.Bbar
    G0 = self.G0
    I2 = self.I2
    beta_0 = self.beta_0
    X1c = self.X1c
    X1s = self.X1s
    Y1c = self.Y1c
    Y1s = self.Y1s
    X20 = self.X20
    X2s = self.X2s
    X2c = self.X2c
    Y20 = self.Y20
    Y2s = self.Y2s
    Y2c = self.Y2c
    Z20 = self.Z20
    Z2s = self.Z2s
    Z2c = self.Z2c
    B20 = self.B20
    B1c = self.B1c
    B1s = self.B1s
    d_l_d_varphi = self.d_l_d_varphi
    torsion = self.torsion
    curvature = self.curvature
    abs_G0_over_B0 = self.abs_G0_over_B0
    d_X1c_d_varphi = self.d_X1c_d_varphi
    d_X1s_d_varphi = self.d_X1s_d_varphi
    d_Y1c_d_varphi = self.d_Y1c_d_varphi
    d_Y1s_d_varphi = self.d_Y1s_d_varphi
    d_Z20_d_varphi = self.d_Z20_d_varphi
    V1 = X1c * X1c + Y1c * Y1c + Y1s * Y1s + X1s * X1s

    if self.omn:
        # The expression below is derived in the O(r^2) paper
        Q = (G2 + I2 * N_helicity) * Bbar * d_l_d_varphi / (2 * G0 * G0) \
                + 2 * (X2c * Y2s - X2s * Y2c) \
                + Bbar / (2*G0) * (d_l_d_varphi * X20 * curvature - d_Z20_d_varphi) \
                + I2 / (4 * G0) * (-d_l_d_varphi * torsion * V1 + Y1c * d_X1c_d_varphi - X1c * d_Y1c_d_varphi + Y1s * d_X1s_d_varphi - X1s * d_Y1s_d_varphi) \
                + beta_0 * Bbar / (4 * G0) * (X1s * d_Y1c_d_varphi + Y1c * d_X1s_d_varphi - X1c * d_Y1s_d_varphi - Y1s * d_X1c_d_varphi)
        flux_constraint_coefficient = - Q * B0 / (2 * sign_G * Bbar)

    else:
        # In quasisymmetry, X1s=0, B1s=0, which simplifies many terms
        # The expression below is computed in "20190305-01 GarrenBoozer r2 corrected radius.nb" in the section "Approach of adding r^3 terms, assuming quasisymmetry"
        # 20190714: To account for QH cases, changed iota -> iota_N where it occurs 3 lines below:
        flux_constraint_coefficient = (-4*B0**2*G0*X20**2*Y1c**2 + 8*B0**2*G0*X20*X2c*Y1c**2 - 4*B0**2*G0*X2c**2*Y1c**2 - \
            4*B0**2*G0*X2s**2*Y1c**2 + 8*B0*G0*B1c*X1c*X2s*Y1c*Y1s + 16*B0**2*G0*X20*X2s*Y1c*Y1s + \
            2*B0**2*I2*self.iotaN*X1c**2*Y1s**2 - G0*B1c**2*X1c**2*Y1s**2 - 4*B0*G0*B20*X1c**2*Y1s**2 - \
            8*B0*G0*B1c*X1c*X20*Y1s**2 - 4*B0**2*G0*X20**2*Y1s**2 - 8*B0*G0*B1c*X1c*X2c*Y1s**2 - \
            8*B0**2*G0*X20*X2c*Y1s**2 - 4*B0**2*G0*X2c**2*Y1s**2 - 4*B0**2*G0*X2s**2*Y1s**2 + \
            8*B0**2*G0*X1c*X20*Y1c*Y20 - 8*B0**2*G0*X1c*X2c*Y1c*Y20 - 8*B0**2*G0*X1c*X2s*Y1s*Y20 - \
            4*B0**2*G0*X1c**2*Y20**2 - 8*B0**2*G0*X1c*X20*Y1c*Y2c + 8*B0**2*G0*X1c*X2c*Y1c*Y2c + \
            24*B0**2*G0*X1c*X2s*Y1s*Y2c + 8*B0**2*G0*X1c**2*Y20*Y2c - 4*B0**2*G0*X1c**2*Y2c**2 + \
            8*B0**2*G0*X1c*X2s*Y1c*Y2s - 8*B0*G0*B1c*X1c**2*Y1s*Y2s - 8*B0**2*G0*X1c*X20*Y1s*Y2s - \
            24*B0**2*G0*X1c*X2c*Y1s*Y2s - 4*B0**2*G0*X1c**2*Y2s**2 - 4*B0**2*G0*X1c**2*Z20**2 - \
            4*B0**2*G0*Y1c**2*Z20**2 - 4*B0**2*G0*Y1s**2*Z20**2 - 4*B0**2*abs_G0_over_B0*I2*Y1c*Y1s*Z2c + \
            8*B0**2*G0*X1c**2*Z20*Z2c + 8*B0**2*G0*Y1c**2*Z20*Z2c - 8*B0**2*G0*Y1s**2*Z20*Z2c - \
            4*B0**2*G0*X1c**2*Z2c**2 - 4*B0**2*G0*Y1c**2*Z2c**2 - 4*B0**2*G0*Y1s**2*Z2c**2 + \
            2*B0**2*abs_G0_over_B0*I2*X1c**2*Z2s + 2*B0**2*abs_G0_over_B0*I2*Y1c**2*Z2s - 2*B0**2*abs_G0_over_B0*I2*Y1s**2*Z2s + \
            16*B0**2*G0*Y1c*Y1s*Z20*Z2s - 4*B0**2*G0*X1c**2*Z2s**2 - 4*B0**2*G0*Y1c**2*Z2s**2 - \
            4*B0**2*G0*Y1s**2*Z2s**2 + B0**2*abs_G0_over_B0*I2*X1c**3*Y1s*torsion + B0**2*abs_G0_over_B0*I2*X1c*Y1c**2*Y1s*torsion + \
            B0**2*abs_G0_over_B0*I2*X1c*Y1s**3*torsion - B0**2*I2*X1c*Y1c*Y1s*d_X1c_d_varphi + \
            B0**2*I2*X1c**2*Y1s*d_Y1c_d_varphi)/(16*B0**2*G0*X1c**2*Y1s**2)

        # The expression below is derived in the O(r^2) paper, and in "20190318-01 Wrick's streamlined Garren-Boozer method, MHD.nb" in the section "Not assuming quasisymmetry".
        # Note Q = (1/2) * (XYEquation0 without X3 and Y3 terms) where XYEquation0 is the quantity in the above notebook.
        Q = -sign_psi * B0 * abs_G0_over_B0 / (2*G0*G0) * (self.iotaN * I2 + mu0 * self.p2 * G0 / (B0 * B0)) + 2 * (X2c * Y2s - X2s * Y2c) \
                + sign_psi * B0 / (2*G0) * (abs_G0_over_B0 * X20 * curvature - d_Z20_d_varphi) \
                + I2 / (4 * G0) * (-abs_G0_over_B0 * torsion * (X1c*X1c + Y1s*Y1s + Y1c*Y1c) + Y1c * d_X1c_d_varphi - X1c * d_Y1c_d_varphi)
        predicted_flux_constraint_coefficient = - Q / (2 * sign_G * sign_psi)

        B0_order_a_squared_to_cancel = -sign_G * B0 * B0 * (G2 + I2 * N_helicity) * abs_G0_over_B0 / (2*G0*G0) \
            -sign_G * sign_psi * B0 * 2 * (X2c * Y2s - X2s * Y2c) \
            -sign_G * B0 * B0 / (2*G0) * (abs_G0_over_B0 * X20 * curvature - d_Z20_d_varphi) \
            -sign_G * sign_psi * B0 * I2 / (4*G0) * (-abs_G0_over_B0 * torsion * (X1c*X1c + Y1c*Y1c + Y1s*Y1s) + Y1c * d_X1c_d_varphi - X1c * d_Y1c_d_varphi)
        self.B0_order_a_squared_to_cancel = B0_order_a_squared_to_cancel

        logger.debug('max|flux_constraint_coefficient - predicted_flux_constraint_coefficient|:',np.max(abs(flux_constraint_coefficient - predicted_flux_constraint_coefficient)))
        logger.debug('max|flux_constraint_coefficient - B0_order_a_squared_to_cancel/(2*B0)|:',np.max(abs(flux_constraint_coefficient - B0_order_a_squared_to_cancel/(2*B0))))

        if np.max(abs(flux_constraint_coefficient - predicted_flux_constraint_coefficient)) > 1e-7 \
        or np.max(abs(flux_constraint_coefficient - B0_order_a_squared_to_cancel/(2*B0))) > 1e-7:
            print("WARNING!!! Methods of computing lambda disagree!! Higher nphi resolution might be needed.")

    self.X3c1 = self.X1c * flux_constraint_coefficient
    self.X3s1 = self.X1s * flux_constraint_coefficient
    self.Y3c1 = self.Y1c * flux_constraint_coefficient
    self.Y3s1 = self.Y1s * flux_constraint_coefficient
    self.Z3c1 = 0
    self.Z3s1 = 0

    self.X3c3 = 0
    self.X3s3 = 0
    self.Y3c3 = 0
    self.Y3s3 = 0
    self.Z3c3 = 0
    self.Z3s3 = 0

    self.d_X3c1_d_varphi = self.d_d_varphi @ self.X3c1
    self.d_X3s1_d_varphi = self.d_d_varphi @ self.X3s1
    self.d_Y3c1_d_varphi = self.d_d_varphi @ self.Y3c1
    self.d_Y3s1_d_varphi = self.d_d_varphi @ self.Y3s1

    self.flux_constraint_coefficient = flux_constraint_coefficient

    if self.helicity == 0:
        self.X3c1_untwisted = self.X3c1
        self.Y3c1_untwisted = self.Y3c1
        self.Y3s1_untwisted = self.Y3s1
        self.X3s1_untwisted = self.X3s1
        self.X3s3_untwisted = self.X3s3
        self.X3c3_untwisted = self.X3c3
        self.Y3c3_untwisted = self.Y3c3
        self.Y3s3_untwisted = self.Y3s3
        self.Z3s1_untwisted = self.Z3s1
        self.Z3s3_untwisted = self.Z3s3
        self.Z3c1_untwisted = self.Z3c1
        self.Z3c3_untwisted = self.Z3c3
    else:
        angle = -self.helicity * self.nfp * self.varphi
        sinangle = np.sin(angle)
        cosangle = np.cos(angle)
        self.X3s1_untwisted = self.X3s1 *   cosangle  + self.X3c1 * sinangle
        self.X3c1_untwisted = self.X3s1 * (-sinangle) + self.X3c1 * cosangle
        self.Y3s1_untwisted = self.Y3s1 *   cosangle  + self.Y3c1 * sinangle
        self.Y3c1_untwisted = self.Y3s1 * (-sinangle) + self.Y3c1 * cosangle
        self.Z3s1_untwisted = self.Z3s1 *   cosangle  + self.Z3c1 * sinangle
        self.Z3c1_untwisted = self.Z3s1 * (-sinangle) + self.Z3c1 * cosangle
        sinangle = np.sin(3*angle)
        cosangle = np.cos(3*angle)
        self.X3s3_untwisted = self.X3s3 *   cosangle  + self.X3c3 * sinangle
        self.X3c3_untwisted = self.X3s3 * (-sinangle) + self.X3c3 * cosangle
        self.Y3s3_untwisted = self.Y3s3 *   cosangle  + self.Y3c3 * sinangle
        self.Y3c3_untwisted = self.Y3s3 * (-sinangle) + self.Y3c3 * cosangle
        self.Z3s3_untwisted = self.Z3s3 *   cosangle  + self.Z3c3 * sinangle
        self.Z3c3_untwisted = self.Z3s3 * (-sinangle) + self.Z3c3 * cosangle
